/**
 * 
 */
package net.javamonkey.syringe;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import net.javamonkey.syringe.models.Argument;
import net.javamonkey.syringe.models.Context;
import net.javamonkey.syringe.models.Dependency;
import net.javamonkey.syringe.models.Factory;
import net.javamonkey.syringe.models.Property;
import net.javamonkey.syringe.utils.type.MethodFinder;
import net.javamonkey.syringe.utils.type.TypeConverter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * @author Matthew Cory
 *
 */
public final class Syringe {
	
	public static void registerContextLoader(final ContextLoader ctxLdr) {
		if (!(contextLoaders.contains(ctxLdr))) {
			contextLoaders.add(ctxLdr);
		}
	}

	public static void loadContexts(final String... contextIdentifiers) {
		 if (contexts == null) {
			contexts = new HashSet<Context>();
		}
		
		for (final String ctxId : contextIdentifiers) {
			final ContextLoader ctxLdr = findContextLoader(ctxId);
			if (ctxLdr == null) {
				log.warn("Could not locate context loader for context {}", ctxId);
				continue;
			}
			
			final Context ctx = ctxLdr.loadContext(ctxId);
			contexts.add(ctx);
		}
	}
	
	public static void clearContexts() {
		if (contexts != null) {
			contexts.clear();
		}
	}

	public static <T> T getItem(final String itemId, final Class<T> klass) {
		Dependency dep = null;
		for (final Context c : contexts) {
			if (c == null) {
				log.warn("Context was null...");
				continue;
			}
			
			dep = c.getDependencyById(itemId);
			if (dep != null) {
				break;
			}
		}
		
		if (dep == null) {
			log.warn("Could not find item with id {}", itemId);
			return null;
		}
		
		if (!(dep.getClassName().equals(klass.getName()))) {
			throw new IllegalArgumentException("Dependency class name (" + dep.getClassName() + ") does not match supplied class name (" + klass.getName() + ").");
		}
		
		T res = null;
		
		if (dep.isSingleton()) {
			if (singletons.containsKey(dep)) {
				return klass.cast(singletons.get(dep));
			}
		}
		
		if (dep.hasFactory()) {
			res = buildByFactory(dep, klass);
		} else if (dep.hasConstructorArgs()) {
			res = buildByConstructorArgs(dep, klass);
		} else {
			try {
				res = klass.newInstance();
			} catch (InstantiationException e) {
				log.error("Error instantiating class.", e);
			} catch (IllegalAccessException e) {
				log.error("Error instantiating class.", e);
			}
		}
		
		try {
			fillProperties(dep, res);
		} catch (IntrospectionException e) {
			log.error("Error filling item properties.", e);
		} catch (IllegalAccessException e) {
			log.error("Error filling item properties.", e);
		} catch (InvocationTargetException e) {
			log.error("Error filling item properties.", e);
		}
		
		// if it's a singleton and we made it this far,
		// it must be the first instance.
		if (dep.isSingleton()) {
			singletons.put(dep, res);
		}
		
		return res;
	}
	
	private static <T> T buildByFactory(final Dependency dep, final Class<T> klass) {
		final Factory factory = dep.getFactory();
		if (factory == null) {
			throw new IllegalStateException("Cannot build via factory method without factory.");
		}
		
		Class<?> factoryClass = null;
		try {
			factoryClass = Class.forName(factory.getClassName());
		} catch (ClassNotFoundException e) {
			log.error("Cannot find specified factory class.", e);
		}
		
		if (factoryClass == null) {
			return null;
		}
		
		final Collection<Argument> factoryArgs = factory.getArguments();
		final Method toInvoke = MethodFinder.findMethod(factoryClass, factory.getMethodName(), factoryArgs);
		
		if (toInvoke == null) {
			return null;
		}
		
		Object[] args = buildFactoryArguments(toInvoke, factoryArgs);
		
		T res = null;
		
		try {
			res = klass.cast(toInvoke.invoke(null, args));
		} catch (IllegalArgumentException e) {
			log.error("Couldn't create by factory", e);
		} catch (IllegalAccessException e) {
			log.error("Couldn't create by factory", e);
		} catch (InvocationTargetException e) {
			log.error("Couldn't create by factory", e);
		}
		
		return res;
	}
	
	private static Object[] buildFactoryArguments(final Method toInvoke, final Collection<Argument> args) {
		final int numArgs = args.size();
		final Object[] res = new Object[numArgs];
		
		final Class<?>[] paramClasses = toInvoke.getParameterTypes();
		for (final Argument arg : args) {
			final int idx = arg.getIndex();
			final Class<?> paramClass = paramClasses[idx];
			res[idx] = TypeConverter.convertString(arg.getStrValue(), paramClass);
		}
		
		return res;
	}
	
	private static <T> T buildByConstructorArgs(final Dependency dep, final Class<T> klass) {
		final Collection<Argument> args = dep.getCtorArguments();
		if ((args == null) || (args.isEmpty())) {
			throw new IllegalStateException("Cannot build via constructor without constructor arguments.");
		}
		
		final Constructor<?>[] ctors = klass.getConstructors();
		Constructor<?> toInvoke = null;
		final int numArgs = args.size();
		for (final Constructor<?> ctor : ctors) {
			final Class<?>[] params = ctor.getParameterTypes();
			if (params.length != numArgs) {
				continue;
			}
			
			// okay, same number or parameters.
			// for now, let's pretend it works as-is.
			toInvoke = ctor;
			break;
		}
		
		final Object[] toPass = buildCtorParameters(toInvoke, new ArrayList<Argument>(args));
		
		T res = null;
		try {
			res = klass.cast(toInvoke.newInstance(toPass));
		} catch (IllegalArgumentException e) {
			log.error("Couldn't create by constructor", e);
		} catch (InstantiationException e) {
			log.error("Couldn't create by constructor", e);
		} catch (IllegalAccessException e) {
			log.error("Couldn't create by constructor", e);
		} catch (InvocationTargetException e) {
			log.error("Couldn't create by constructor", e);
		}
		
		return res;
	}
	
	private static Object[] buildCtorParameters(Constructor<?> toInvoke, List<Argument> args) {
		final Class<?>[] params = toInvoke.getParameterTypes();
		
		final int numParams = params.length;
		final Object[] res = new Object[numParams];
		for (int i = 0; i < numParams; ++i) {
			final Argument arg = args.get(i);
			final Class<?> paramClass = params[i];
			final int idx = arg.getIndex();
			
			res[idx] = TypeConverter.convertString(arg.getStrValue(), paramClass);
		}
		
		return res;
	}

	private static <T> void fillProperties(final Dependency dep, final T item) throws IntrospectionException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		final Collection<Property> props = dep.getProperties();
		if ((props == null) || (props.isEmpty())) {
			return;
		}
		
		final BeanInfo bi = Introspector.getBeanInfo(item.getClass());
		final PropertyDescriptor[] propDescs = bi.getPropertyDescriptors();
		
		for (final Property p : props) {
			for (final PropertyDescriptor pd : propDescs) {
				if (!(pd.getName().equals(p.getName()))) {
					continue;
				}
				
				final Method m = pd.getWriteMethod();
				final Class<?> propType = pd.getPropertyType();
				final Object propValue = convertString(p.getStrValue(), propType);
				m.invoke(item, propValue);
			}
		}
	}
	
	private static Object convertString(final String toConv, final Class<?> destType) {
		final Object res;
		if (destType == String.class) {
			res = toConv;
		} else if (TypeConverter.hasConverter(destType)) {
			res = TypeConverter.convertString(toConv, destType);
		} else {
			log.warn("Didn''t recognize type: {}", destType.getCanonicalName());
			res = null;
		}
		
		return res;
	}

	private static ContextLoader findContextLoader(final String ctxId) {
		for (final ContextLoader ldr : contextLoaders) {
			if (ldr.canLoad(ctxId)) {
				return ldr;
			}
		}
		
		return null;
	}
	
	private static final Logger log = LoggerFactory.getLogger(Syringe.class);
	private static Set<Context> contexts;
	private static Set<ContextLoader> contextLoaders = new HashSet<ContextLoader>();
	private static Map<Dependency, Object> singletons = new HashMap<Dependency, Object>();

	private Syringe() {
	
	}
	
}
