/**
 * 
 */
package net.javamonkey.syringe.utils.type;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Matthew Cory
 *
 */
public final class TypeConverter {
	
	private static final Logger log = LoggerFactory.getLogger(TypeConverter.class);
	
	public static <T> void registerConverter(final Class<T> klass, final ObjectConverter converter) {
		converters.put(klass, converter);
	}
	
	public static Object convertString(final String value, final Class<?> klass) {
		final Object res;
		if (klass == String.class) {
			res = value;
		} else if (PrimitiveConverter.isPrimitive(klass)) {
			res = convertPrimitive(value, klass);
		} else {
			res = converters.get(klass).convertString(value);
		}
		
		return res;
	}

	public static boolean canConvert(String strValue, Class<?> destType) {
		if (!hasConverter(destType)) {
			return false;
		}
		
		return (convertString(strValue, destType) != null);
	}

	public static boolean hasConverter(Class<?> destType) {
		return (destType == String.class ||
				converters.containsKey(destType) || 
				primitiveConverters.containsKey(destType));
	}

	private static Map<Class<?>, ObjectConverter> initializeConverters() {
		final Map<Class<?>, ObjectConverter> res = new HashMap<Class<?>, ObjectConverter>();
		
		res.put(int.class, new ObjectConverter() {
			@Override
			public Object convertString(String value) {
				return PrimitiveConverter.stringToIntPrim(value);
			}
		});
		
		res.put(Integer.class, null);
		
		return res;
	}

	private static Object convertPrimitive(String value, Class<?> klass) {
		final Method m = primitiveConverters.get(klass);
		if (m == null) {
			log.error("Could not locate converter for " + klass.getCanonicalName());
			return null;
		}
		
		Object res = null;
		try {
			res = m.invoke(PrimitiveConverter.class, value);
		} catch (IllegalArgumentException e) {
			log.error(e.getMessage(), e);
		} catch (IllegalAccessException e) {
			log.error(e.getMessage(), e);
		} catch (InvocationTargetException e) {
			log.error(e.getMessage(), e);
		}
		
		return res;
	}
	
	private static final Map<Class<?>, ObjectConverter> converters = initializeConverters();
	private static final Map<Class<?>, Method> primitiveConverters = initializePrimitiveConverters();

	private static Map<Class<?>, Method> initializePrimitiveConverters() {
		final Map<Class<?>, Method> res = new HashMap<Class<?>, Method>();
		
		try {
			final Method[] methods = PrimitiveConverter.class.getMethods();
			for (final Method m : methods) {
				if (!m.isAnnotationPresent(TypeConverterMethod.class)) {
					continue;
				}
				
				final TypeConverterMethod typeConverter = m.getAnnotation(TypeConverterMethod.class);
				res.put(typeConverter.value(), m);
			}
		} catch (SecurityException e) {
						log.error(e.getMessage(), e);
		}
		
		return res;
	}

	private TypeConverter() {
	}
}
