/**
 * 
 */
package net.javamonkey.syringe.xml;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import net.javamonkey.syringe.ContextLoader;
import net.javamonkey.syringe.models.Context;
import net.javamonkey.syringe.models.Dependency;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * @author Matthew Cory
 *
 */
public final class XmlContextLoader implements ContextLoader {

	private static final Logger log = LoggerFactory.getLogger(XmlContextLoader.class);
	
	/* (non-Javadoc)
	 * @see net.javamonkey.syringe.ContextLoader#loadContext(java.lang.String[])
	 */
	@Override
	public Context loadContext(final String contextId) {
		final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder db = null;
		Document doc = null;
		try {
			db = factory.newDocumentBuilder();
			doc = db.parse(contextId);
		} catch (ParserConfigurationException pce) {
			log.error("Error configuring XML parser.", pce);
		} catch (SAXException saxe) {
			log.error("Error parsing document.", saxe);
		} catch (IOException ioe) {
			log.error("Error loading XML file.", ioe);
		}
		
		if ((db == null) || (doc == null)) {
			log.warn("An error occured while initializing XML document; no context will be loaded.");
			return null;
		}
		
		final Context res = new Context();
		
		final Collection<Dependency> dependencies = buildDependencies(doc);
		log.info("Number of dependencies: {}", dependencies.size());
		res.setDependencies(dependencies);
		
		return res;
	}

	@Override
	public boolean canLoad(final String ctxId) {
		return (ctxId.toLowerCase().endsWith(".xml"));
	}

	private Collection<Dependency> buildDependencies(final Document doc) {
		final Element root = doc.getDocumentElement();
		final NodeList depNodes = root.getChildNodes();
		final int depNodesLength = depNodes.getLength();
		
		final Collection<Dependency> res = new ArrayList<Dependency>(depNodesLength);
		for (int i = 0; i < depNodesLength; ++i) {
			final Node node = depNodes.item(i);
			
			if (!node.getNodeName().toLowerCase().contains("dependency")) {
				continue;
			}
			
			final Dependency dep = buildDependency(node);
			res.add(dep);
		}
		
		return res;
	}
	
	private Dependency buildDependency(final Node node) {
		final Dependency res = new Dependency();
		
		final NamedNodeMap attrs = node.getAttributes();
		
		final String id = attrs.getNamedItem("id").getNodeValue();
		final String className = attrs.getNamedItem("class").getNodeValue();
		
		final Node singletonNode = attrs.getNamedItem("singleton");
		if (singletonNode != null) {
			res.setSingleton(singletonNode.getNodeValue().toLowerCase().equals("true"));
		}
		
		res.setId(id);
		res.setClassName(className);
		
		final NodeList depChildren = node.getChildNodes();
		buildChildren(res, depChildren);
		return res;
	}

	private void buildChildren(final Dependency res, final NodeList depChildren) {
		if ((depChildren == null) || (depChildren.getLength() == 0)) {
			return;
		}
		
		final int depChildLength = depChildren.getLength();
		for (int i = 0; i < depChildLength; ++i) {
			final Node node = depChildren.item(i);
			if (!(node instanceof Element)) {
				continue;
			}
			
			final String nodeName = node.getNodeName();
			final Element elem = (Element)node;
			
			final XmlChildBuilder builder = childBuilders.get(nodeName);
			if (builder == null) {
				log.warn("No builder for node {}", nodeName);
				continue;
			}
			
			builder.buildChild(res, elem);
		}
	}
	
	private static Map<String, XmlChildBuilder> initializeChildBuilders() {
		final Map<String, XmlChildBuilder> res = new HashMap<String, XmlChildBuilder>();
		
		res.put("property", new XmlPropertyBuilder());
		res.put("ctor-arg", new XmlCtorArgumentBuilder());
		res.put("factory", new XmlFactoryBuilder());
		
		return res;
	}
	
	private static Map<String, XmlChildBuilder> childBuilders = initializeChildBuilders();
}
